#include "squidwall.h"

/* every 3 connections to clamav, we are checking the version of it */
#define CLAMD_VERSION_CHECK 3

/* is someone working on such a thing ? */
#define HAVE_PATCHED_SQUID  0

/* buffers for clamd command connection */
static char clamd_iBUF[BUFSIZE]; buffer clamd_in;
static char clamd_oBUF[BUFSIZE]; buffer clamd_out;

/* buffers for clamd data/stream connection */
static char clams_oBUF[BUFSIZE]; buffer clams_out;

/* buffers for proxy connection */
static char proxy_iBUF[BUFSIZE]; buffer proxy_in;
static char proxy_oBUF[BUFSIZE]; buffer proxy_out;

stralloc virus;
stralloc clamav;
stralloc filter;

int check_cont(void)
{
	struct squidwall_string *s;
	struct squidwall_regex *r;
	int64 fd_clamd, fd_stream, fd_proxy;
	static int clamd_check_version=CLAMD_VERSION_CHECK;
	size_t len=0;
	char ch;
	int i;
	stralloc clamdtext;
	stralloc proxytext;
	stralloc content;

	if (!str_equal(squid_method, "GET"))
		return ACL_ALLOW;

	stralloc_init(&content);
	stralloc_init(&clamdtext);
	stralloc_init(&proxytext);
	
	/* for closing all connection, before returning */
	void close_connections(void)
	{
		buffer_flush(&clams_out);

		/* close proxy and clamd stream connection */
		clams_out.fd=-1; buffer_close(&clams_out); io_close(fd_stream);
		proxy_out.fd=-1; buffer_close(&proxy_out); io_close(fd_proxy);
		stralloc_free(&proxytext);

		/* see, what clamd means to our stream */
		io_wantread(fd_clamd); io_wait(); io_dontwantread(fd_clamd);
		if (!stralloc_copys(&clamdtext, "")) die_nomem();
		while (buffer_getline_sa(&clamd_in, &clamdtext) != 0);
	
		/* close clamd connection */
		clamd_out.fd=-1; buffer_close(&clamd_out); io_close(fd_clamd);
		stralloc_free(&clamdtext);
	}

	/* get version of clamav */
	if (clamd_check_version % CLAMD_VERSION_CHECK == 0) {
		debug(9, "check_cont(), reading clamd version...");

		fd_clamd=socket_tcp4();
		if (fd_clamd == -1) return panic("socket() for clamd failed");
		i=socket_connect4(fd_clamd, (char*)&clamd_ip, clamd_port);

		if (i == -1) return panic("connect() to clamd failed");

		if (!io_fd(fd_clamd)) return panic("io_fd() failed");

		buffer_init(&clamd_in, (int(*)())read, fd_clamd, clamd_iBUF,
			sizeof clamd_iBUF);
		buffer_init(&clamd_out, (int(*)())write, fd_clamd, clamd_oBUF,
			sizeof clamd_oBUF);

		io_wantwrite(fd_clamd); io_wait(); io_dontwantwrite(fd_clamd);
		buffer_putsflush(&clamd_out, "VERSION\n");

		io_wantread(fd_clamd); io_wait(); io_dontwantread(fd_clamd);
		if (buffer_getnewline_sa(&clamd_in, &clamav) == 0)
			return panic("can't read response of clamd");
		clamav.s[clamav.len]=0;

		if (!str_start(clamav.s, "ClamAV "))
			return panic("bad response from clamd, expected: 'ClamAV ' ...");
	
		clamd_in.fd=-1;  buffer_close(&clamd_in);
		clamd_out.fd=-1; buffer_close(&clamd_out);
		io_close(fd_clamd);
		clamd_check_version=0;
	}
	clamd_check_version++;

	/* connection to clamd */
	fd_clamd=socket_tcp4();
	if (fd_clamd == -1) return panic("socket() for clamd failed");
        i=socket_connect4(fd_clamd, (char*)&clamd_ip, clamd_port);
	if (i == -1) return panic("connect() to clamd failed");
	if (!io_fd(fd_clamd)) return panic("io_fd() failed");

	buffer_init(&clamd_in,   (int(*)())read, fd_clamd, clamd_iBUF, sizeof clamd_iBUF);
	buffer_init(&clamd_out, (int(*)())write, fd_clamd, clamd_oBUF, sizeof clamd_oBUF);

	io_wantwrite(fd_clamd); io_wait(); io_dontwantwrite(fd_clamd);
	buffer_putsflush(&clamd_out, "STREAM\n");

	io_wantread(fd_clamd); io_wait(); io_dontwantread(fd_clamd);
	if (buffer_getnewline_sa(&clamd_in, &clamdtext) == 0)
		return panic("can't read response of clamd");
	clamdtext.s[clamdtext.len]=0;
	if (!str_start(clamdtext.s, "PORT "))
		return panic("bad response from clamd, expected: PORT xxx");

	i=scan_uint(clamdtext.s+5, &clams_port);
	if (i == 0) panic("can't read port of clamd stream");

	/* open connection to clamd stream socket */
	fd_stream=socket_tcp4();
	if (fd_stream == -1) return panic("socket() for clamd failed");
        i=socket_connect4(fd_stream, (char*)&clamd_ip, clams_port);
	if (i == -1) return panic("connect() to clamd failed");
	buffer_init(&clams_out, (int(*)())write, fd_stream, clams_oBUF, sizeof clams_oBUF);
	if (!io_fd(fd_stream)) panic("io_fd() failed");

	/* open connection to proxy */
	fd_proxy=socket_tcp4();
	if (fd_stream == -1) return panic("socket() for proxy failed");
        i=socket_connect4(fd_proxy, (char*)&proxy_ip, proxy_port);
	if (i == -1) return panic("connect() to proxy failed");
	if (!io_fd(fd_proxy)) return panic("io_fd() failed");

	buffer_init(&proxy_in,   (int(*)())read, fd_proxy, proxy_iBUF, sizeof proxy_iBUF);
	buffer_init(&proxy_out, (int(*)())write, fd_proxy, proxy_oBUF, sizeof proxy_oBUF);

	io_wantwrite(fd_proxy); io_wait(); io_dontwantwrite(fd_proxy);
	buffer_puts(&proxy_out, squid_method);
	buffer_puts(&proxy_out, " ");
	buffer_puts(&proxy_out, squid_url);
	buffer_putsflush(&proxy_out, " HTTP/1.0\r\n\r\n");

	while (buffer_get(&proxy_in, &ch, 1) == 1) {
		buffer_put(&clams_out, &ch, 1);
		if (++len >= maxsize) break;
#if HAVE_PATCHED_SQUID
		/* - can be used, if squid understands a '*' for sending
		 *   some first data to the client
		 * - should become an extension the the redirector interface!
		 *   /TR
		 */
		if (len % BUFSIZE == 0) buffer_putsflush(buffer_1, "*");
#endif
		/* content scanning with regex.7 ;) */
		if (!stralloc_catb(&content, &ch, 1)) die_nomem();
		if (ch == '\n' || ch == '\r') {
			content.s[content.len]=0;

			if (cfilter_user)
			for (i=0; i<array_length(cfilter_user, sizeof(struct squidwall_string)); i++) {
				s = array_get(cfilter_user, sizeof(struct squidwall_string), i);
				if (case_strstr(content.s, s->string.s)) {
					close_connections();
					stralloc_copy(&filter, &s->name);
					redir_site(REDIR_TYPE_CONT);
					return ACL_DENY;
				}
			}

			if (crfilter_user)
			for (i=0; i<array_length(crfilter_user, sizeof(struct squidwall_regex)); i++) {
				r = array_get(crfilter_user, sizeof(struct squidwall_regex), i);
				if (regexec(&r->regex, content.s, 0, 0, 0) == 0) {
					close_connections();
					stralloc_copy(&filter, &r->name);
					redir_site(REDIR_TYPE_CONT);
					return ACL_DENY;
				}
			}

			if (cfilter_host)
			for (i=0; i<array_length(cfilter_host, sizeof(struct squidwall_string)); i++) {
				s = array_get(cfilter_host, sizeof(struct squidwall_string), i);
				if (case_strstr(content.s, s->string.s)) {
					close_connections();
					stralloc_copy(&filter, &s->name);
					redir_site(REDIR_TYPE_CONT);
					return ACL_DENY;
				}
			}

			if (crfilter_host)
			for (i=0; i<array_length(crfilter_host, sizeof(struct squidwall_regex)); i++) {
				r = array_get(crfilter_host, sizeof(struct squidwall_regex), i);
				if (regexec(&r->regex, content.s, 0, 0, 0) == 0) {
					close_connections();
					stralloc_copy(&filter, &r->name);
					redir_site(REDIR_TYPE_CONT);
					return ACL_DENY;
				}
			}

			stralloc_copys(&content, "");
		}
	}
	buffer_flush(&clams_out);

	/* close proxy and clamd stream connection */
	clams_out.fd=-1; buffer_close(&clams_out); io_close(fd_stream);
	proxy_out.fd=-1; buffer_close(&proxy_out); io_close(fd_proxy);
	stralloc_free(&proxytext);

	/* see, what clamd means to our stream */
	io_wantread(fd_clamd); io_wait(); io_dontwantread(fd_clamd);
	if (!stralloc_copys(&clamdtext, "")) die_nomem();
	while (buffer_getline_sa(&clamd_in, &clamdtext) != 0);
	
	/* close clamd connection */
	clamd_out.fd=-1; buffer_close(&clamd_out); io_close(fd_clamd);

	if (clamdtext.len != 11) { /* "stream: OK\n" -> if len==11 => no virus! */
		/* "stream: Worm.Sober.I FOUND\n" -> "Worm.Sober.I" */
		clamdtext.s[clamdtext.len-7]=0; /* remove " FOUND\n" */
		if (!stralloc_copys(&virus, clamdtext.s+8)) die_nomem();
		stralloc_free(&clamdtext);

		/* log the virus site */
		log_headline("virus");
		log_sa(&url);    log_s("\t");
		log_sa(&ip);     log_s("\t");
		log_sa(&host);   log_s("\t");
		log_sa(&user);   log_s("\t");
		log_sa(&method); log_s("\t");
		log_sa(&virus);  log_s("\n");

		redir_site(REDIR_TYPE_VIRUS);
		return ACL_DENY;
	}
	stralloc_free(&clamdtext);

	return ACL_ALLOW;
}
