#ifndef SW_SQUIDWALL_H
#define SW_SQUIDWALL_H

#include "params.h"

#define VERSION    "squidwall 0.2a"
#define BUFSIZE    5000

/* system */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <regex.h>
#include <fcntl.h>
#include <errno.h>
#include <wait.h>
#include <time.h>

/* libowfat */
#include "array.h"
#include "buffer.h"
#include "case.h"
#include "fmt.h"
#include "ip4.h"
#include "ip6.h"
#include "io.h"
#include "str.h"
#include "scan.h"
#include "socket.h"
#include "stralloc.h"
#include "textcode.h"

/* squidwall */
#include "debug.h"
#include "env.h"
#include "log.h"

struct ip_address { unsigned char d[4]; };

/* acl rules */
#define ACL_ALLOW    0
#define ACL_FILTER   1
#define ACL_DENY     2
#define ACL_DEFAULT  ACL_ALLOW

/* todo flags for check_cont() */
#define TODO_NONE     0
#define TODO_URL      1
#define TODO_URLr     2
#define TODO_VIRUS    4
#define TODO_CONT     8
#define TODO_CONTr   16
#define TODO_DEFAULT (TODO_URL|TODO_URLr|TODO_VIRUS|TODO_CONT|TODO_CONTr)

/* user / host */
#define TYPE_USER  0
#define TYPE_HOST  1

struct squidwall_acl {
	stralloc name;  /* filename: 'ip|host' or 'user' */
	int flags:3;    /* ACL_(ALLOW|FILTER|DENY) */
	int todo:5;     /* TODO_(NONE|VIRUS|CONT) */
	array cfilter;  /* array for content filtering */
	array ufilter;  /* array for url filtering */
	array crfilter; /* array for regex content filtering */
	array urfilter; /* array for regex url filtering */
};

struct squidwall_regex {
	stralloc name; /* filtername: 'file / regexpression' */
	regex_t regex;
};

struct squidwall_string {
	stralloc name; /* filtername: 'file / string' */
	stralloc string;
};

/* fd of logfile */
extern int64 fd_log;

/* ip's and ports */
extern struct ip_address clamd_ip;
extern struct ip_address proxy_ip;
extern unsigned clamd_port;
extern unsigned clams_port;
extern unsigned proxy_port;

/* max size for content/virus scanning */
extern unsigned long maxsize;

/* pointer to the infos from squid @ fd=0 */
extern char *squid_url;
extern char *squid_ip;
extern char *squid_host;
extern char *squid_user;
extern char *squid_method;

/* filters, of current user and host */
extern array *cfilter_host;
extern array *ufilter_host;
extern array *crfilter_host;
extern array *urfilter_host;

extern array *cfilter_user;
extern array *ufilter_user;
extern array *crfilter_user;
extern array *urfilter_user;

/* destination websites for squid */
extern char *redir_badurl;
extern char *redir_badcont;
extern char *redir_baduser;
extern char *redir_badhost;
extern char *redir_virusfound;

/* misc directories */
extern char *hosts_dir;
extern char *users_dir;
extern char *cfilters_dir;
extern char *ufilters_dir;
extern char *crfilters_dir;
extern char *urfilters_dir;

/* decoded infos */
extern stralloc ip;
extern stralloc url;
extern stralloc host;
extern stralloc user;
extern stralloc method;
extern stralloc redirsite;

/* panic.c */
extern int panic(char *msg);

/* check_acl.c */
extern int check_acl(int type);

/* check_url.c */
extern int check_url(void);

/* check_clamd.c */
extern int check_cont(void);

/* case_strstr.c -> 1=>found | 0=>not found */
extern int case_strstr(const char *haystack, const char *needle);

/* redir_site.c */
#define REDIR_TYPE_USER  0 /* user not okay */
#define REDIR_TYPE_HOST  1 /* host not okay */
#define REDIR_TYPE_URL   2 /* url not okay */
#define REDIR_TYPE_CONT  4 /* content not okay */
#define REDIR_TYPE_VIRUS 8 /* virus found */
extern void redir_site(int type);

#define isset(v,x)     (((v)&(x)) == (x))
#define set(v,x)       (v) |=  (x)
#define unset(v,x)     (v) &= ~(x)

/* read_acls.c */
extern int read_acls(array *root, const char *path, time_t *mtime);

/* read_filters.c */
extern void read_filters(struct squidwall_acl *acl, const char *path, const char *file);

/* read_optionline.c */
extern void read_optionline(struct squidwall_acl *acl, int todo, const char *beginning);

#endif
